"""
directory_builder
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`DirectoryBuilder` along with supporting classes,
functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import errno
import os
from abc import ABC, abstractmethod

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.archetype_resource_builder import ArchetypeResourceBuilder
from inception_tools.constants import UNIMPLEMENTED_ABSTRACT_METHOD_ERROR


class DirectoryBuilder(ArchetypeResourceBuilder, ABC):
    """
    An :py:class:`inception_tools.ArchetypeResourceBuilder` implementation which
    builds an empty directory at it's :py:meth:`path`.  This class provides a
    'template-method' implementation of :py:meth:`path` and :py:meth`build`,
    which delegate responsibility for determining the subpath under ``root_dir`` to a
    method called :py:class:`subpath`.
    """

    def _path(self, root_dir: str, params: ArchetypeParameters):
        return os.path.join(root_dir, self.subpath(params))

    def path(self, root_dir: str, params: ArchetypeParameters) -> str:
        """
        Returns the path to the directory that will be created by
        :py:meth:`build`.
        """
        return self._path(root_dir, params)

    def build(self, root_dir: str, params: ArchetypeParameters) -> None:
        """
        Creates a directory under ``root_dir`` determined by the _result of
        :py:meth:`subpath`.
        :param root_dir: the root directory to which :py:meth:`subpath`
        should be appended
        :param params: the parameters used to determine the subpath
        :return: :py:const:`None`
        ..:note: this method has dependencies on the implementations of
        :py:meth:`subpath`
        """
        p = self._path(root_dir, params)
        try:
            os.makedirs(p)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

    @abstractmethod
    def subpath(self, params: ArchetypeParameters) -> str:
        """
        Subclasses are required to implement this method.  Implementations of this
        method should return the subpath, under the root directory, of the directory
        to be created by :py:meth:`build`.
        :param params: the :py:class `ArchetypeParameters` to use as context
        when creating the subpath, e.g., when creating directories whose
        sub-path might be determined by the package name.
        :return: the path
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR



"""
constants
~~~~~~~~~

Houses the declaration of constants used throughout the :py:mod:`inception_tools`
package. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

UNIMPLEMENTED_ABSTRACT_METHOD_ERROR = NotImplementedError(
    "Method to be implemented by subclasses."
)
"""
The error raised by unimplemented abstract methods declared in :py:class:`ABC`
subclasses.
"""



"""
template_archetype
~~~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`TemplateArchetype` along with
supporting classes, functions, and attributes.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import os

from inception_tools.archetype_base import ArchetypeBase
from inception_tools.archetype_descriptor import ArchetypeDescriptor
from inception_tools.archetype_metadata import ArchetypeMetadata
from inception_tools.template_directory_builder import TemplateDirectoryBuilder
from inception_tools.template_file_builder import TemplateFileBuilder


class TemplateArchetype(ArchetypeBase):
    """
    An :py:class:`Archetype` subclass that uses a set of :py:class:`jinja2` template
    files and descriptor files, all contained within a single directory, to create
    and initial project structure.  At a minimum the directory must contain two JSON
    files

    - :py:attr:`METADATA_FILE_NAME` and
    _ :py:attr:`DESCRIPTOR_FILE_NAME`

    any 'prototype' template files referenced by the descriptor JSON must also be
    present in the directory, at the subpath specified by the descriptor.

    Variables available to the template are the fields of
    :py:class:`ArchetypeParameters`.  Each field of :py:class:`Archetype` parameters
    is available as an unscoped variable.  For example, to reference the fields
    :py:attr:`ArchetypeParameters.author`, you would simply use the variable name
    'author' directly using double curly braces like so: ``{{author}}``.
    """

    METADATA_FILE_NAME = "archetype-metadata.json"
    """
    The name of the JSON file containing :py:class:`ArchetypeMetadata`. This file is
    used to determine what the canonical name of the archetype.
    """

    DESCRIPTOR_FILE_NAME = "archetype-descriptor.json"
    """
    The name of the JSON file containing :py:class:`ArchetypeDescriptor` data. This
    file contains all of the information for mapping template files to sub-paths
    within the projects created by ``inception_tools``.
    """

    def __init__(self, dir_path: str) -> None:
        """
        Initializes a new :py:class:`Archetype` instance from an directory assumed to
        contain an :py:class:`ArchetypeMetadata` JSON file,
        a :py:class:`ArchetypeDescriptor` file, and any prototypes referenced through
        the descriptor.

        :param dir_path: the path to the directory containing the :py:class:`Archetype`
        files
        """
        self._dir_path = dir_path

        with open(self._metadata_path(dir_path)) as metadata_f:
            metadata = ArchetypeMetadata.from_text_io(metadata_f)
        self._metadata = metadata

        with open(self._descriptor_path(dir_path)) as descriptor_f:
            descriptor = ArchetypeDescriptor.from_text_io(descriptor_f)
        self._descriptor = descriptor

        file_builders = self._get_file_builders(dir_path, descriptor)
        dir_builders = self._get_directory_builders(dir_path, descriptor)

        super().__init__(file_builders, dir_builders)

    @classmethod
    def _get_file_builders(cls, dir_path, descriptor):

        file_builders = []
        for f in descriptor.files:
            f_path = os.path.join(dir_path, f.prototype)
            with open(f_path) as fin:
                prototype_content = fin.read()
            fb = TemplateFileBuilder.from_strings(f.subpath, prototype_content)
            file_builders.append(fb)

        return tuple(file_builders)

    @classmethod
    def _get_directory_builders(cls, dir_path, descriptor):
        result = []
        for d in descriptor.directories:
            dd = TemplateDirectoryBuilder.from_string(d.subpath)
            result.append(dd)

        return tuple(result)

    @classmethod
    def _metadata_path(cls, dir_path):
        return os.path.join(dir_path, cls.METADATA_FILE_NAME)

    @classmethod
    def _descriptor_path(cls, dir_path):
        return os.path.join(dir_path, cls.DESCRIPTOR_FILE_NAME)



"""
exception
~~~~~~~~~

Houses the custom:py:class:`Exception` classes used by the system.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2022 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"


class LoggingConfigError(Exception):
    """
    An exception used to signal an error with the logging configuration.
    """



"""
resource_builder
~~~~~~~~~~~~~~~~

Houses the declaration of :py:class:`ArchetypeResourceBuilder` along with
supporting classes, functions, and attributes.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from abc import ABC

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.constants import UNIMPLEMENTED_ABSTRACT_METHOD_ERROR


class ArchetypeResourceBuilder(ABC):
    """
    Provides the common interface by which all
    :py:class:`inception_tools.Archetype` resources are created.
    """

    def path(self, root_dir: str, params: ArchetypeParameters) -> str:
        """
        Returns the full path (possibly non-absolute) to the resource that will be
        created by :py:meth:`build`.
        :param root_dir: the root directory argument for :py:meth:`build`
        :param params: the parameters argument for :py:meth:`build`
        :return: the path
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    def build(self, root_dir: str, params: ArchetypeParameters) -> None:
        """
        Builds the resource and stores it under the root directory.  The content of
        the resource as well as the subpath under the root directory may be,
        but need-not be, determined by the parameters argument.
        :param root_dir: the root directory under which the resource
        should be stored
        :param params: the parameters used to determine the resource's
        content and possibly the subpath under the root directory
        :return: :py:const:`None`
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR



"""
archetype
~~~~~~~~~

Houses the declaration of :py:class:`archetype` along with supporting classes,
functions, and attributes. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

from abc import ABC, abstractmethod
from typing import Iterable

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.constants import UNIMPLEMENTED_ABSTRACT_METHOD_ERROR


class Archetype(ABC):
    """
    Instances of this class provide a canonical template project structure and are
    capable of building particular project structures under a specified root
    directory using a set of parameters to determine the specifics of directory
    structure and file paths and content.
    """

    @abstractmethod
    def file_paths(self, root_path: str, params: ArchetypeParameters) -> Iterable[str]:
        """
        The paths of every file that will be created by :py:meth:`Archetype.build`.

        :param root_path: the root directory of the project structure to be
        created
        :param params: the :py:class:`ArchetypeParameters` to use as context
        for the project to be built
        :return: :py:const:`None`
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    @abstractmethod
    def dir_paths(self, root_path: str, params: ArchetypeParameters) -> Iterable[str]:
        """
        The paths of every directory that will be explicitly created by
        :py:meth:`Archetype.build`.

        :param root_path: the root directory of the project structure to be
        created
        :param params: the :py:class:`ArchetypeParameters` to use as context
        for the project to be built
        :return: :py:const:`None`
        .. note::
           Intermediary directories in the paths listed by
           :py:meth:`file_paths` may also be create even though they are not
           listed under this method.
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    @abstractmethod
    def build(self, root_dir: str, params: ArchetypeParameters) -> None:
        """
        Builds the project structure for this instance.  See the class-level
        documentation of :py:class:`Archetype` for more information.

        :param root_dir: the root directory of the project structure to be
        created
        :param params: the :py:class:`ArchetypeParameters` to use as context
        for the project to be built
        :return: :py:const:`None`
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR



"""
main
~~~~~

Main entry point commend line script.  See :py:func:`cli` for details of how to
invoke this script. """

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import logging
import pathlib
from datetime import datetime
from logging.config import fileConfig

import click

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.exception import LoggingConfigError
from inception_tools.standard_archetype import StandardArchetype


def _logger():
    return logging.getLogger(__file__)


def _incept(package_name, author, author_email):
    params = ArchetypeParameters(
        package_name=package_name,
        author=author,
        author_email=author_email,
        date=datetime.now(),
    )
    archetype = StandardArchetype.CLI
    archetype.build(root_dir=package_name, params=params)


@click.command()
@click.argument("package_name")
@click.argument("author")
@click.argument("author_email")
def incept(package_name, author, author_email):
    """
    Builds a new project structure with the given package name.  Command line
    syntax:

        it incept <package-name> <author-name> <author-email>

    Invoking the command line above will _result in the creation of a directory with
    the following structure:

    \b
        <working-dir>/
            my_package/
                my_package/
                    __init__.py
                    my_package.py
                tests/
                    __init__.py
                    end-to-end/
                        __init__.py
                        test_my_package/
                            __init__.py
                    integration/
                        __init__.py
                        test_my_package/
                            __init__.py
                    unit/
                        __init__.py
                        test_my_package/
                            __init__.py
                LICENSE
                Makefile
                Pipfile
                README.rst
                setup.cfg
                setup.py

    where many of the files are parameterize with the package name, author name,
    author email, etc.

    PACKAGE_NAME: the name of the package to be incepted.  This should be the package
    name, as you would expect it to appear in code references ( e.g. 'my_package' and
    not 'my-package'

    AUTHOR: the name of the package author.  This string is used in copyright
    notices, setup.py package metadata, and for __author__ assignments in Python stub
    files.

    AUTHOR_EMAIL: the email of the package author.  This is used in setup.py package
    metadata and in the auto-generated boiler-plate text of README.rst file.
    """
    try:
        _incept(package_name, author, author_email)
    except Exception:
        msg = (
            "Unexpected exception: "
            f"package_name={package_name}, author={author}, author_email={author_email}"
        )
        _logger().exception(msg)
        raise


@click.group()
@click.option(
    "-l", "--logging-config", default=None, type=click.Path(exists=True, dir_okay=False)
)
def cli(logging_config):
    """
    Main command-line application for the inception_tools package.  This application
    can be used to access various commands listed below.  For example to incept a new
    project called 'my_package', use the following command:

        it incept <package-name> <author-name> <author-email>

    For additional help using any command, use the help for the command as follows

        it <command> --help
    """
    logging_config = logging_config or "log.cfg"
    if pathlib.Path(logging_config).is_file():
        try:
            fileConfig(logging_config, disable_existing_loggers=False)
        except Exception:
            raise LoggingConfigError(
                f"Could not parse the logging config file: {logging_config!r}"
            )
    else:
        logging.basicConfig(level=logging.INFO)
        _logger().warning(
            f"Logging config file ({logging_config!r}) not found.  "
            "All logs will be directed to a basic console logger."
        )
    _logger().debug("Successfully set up logging.")


cli.add_command(incept)

if __name__ == "__main__":
    cli()



"""
file_builder
~~~~~~~~~~~~~

Houses the declaration of :py:class:`FileBuilder` along with supporting
classes, functions, and attributes.
"""

__author__ = "Andrew van Herick"
__copyright__ = "Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved."
__license__ = "Apache Software License 2.0"

import errno
import os
from abc import ABC, abstractmethod

from inception_tools.archetype_parameters import ArchetypeParameters
from inception_tools.archetype_resource_builder import ArchetypeResourceBuilder
from inception_tools.constants import UNIMPLEMENTED_ABSTRACT_METHOD_ERROR


class FileBuilder(ArchetypeResourceBuilder, ABC):
    """
    This abstract base provides a common interface for building files for an an
    :py:class:`inception_tools.Archetype`.  It provides 'template-method'
    implementations of the methods, :py:meth:`path` and:py:method:`build` which rely
    on subclass implementations to determine a subpath under the project root
    directory, as well as how to render file content, both from a
    :py:class:`inception_tools.ArchetypeParameters`.
    """

    def _path(self, root_dir, params):
        return os.path.join(root_dir, self.subpath(params))

    def path(self, root_dir: str, params: ArchetypeParameters) -> str:
        """
        Uses the implementation of :py:meth:`subpath` to determine the absolute path
        to the file saved by :py:meth:`build`.
        :param root_dir: the root directory argument for
        :py:meth:`build`
        :param params: the parameters argument for :py:meth:`build`
        :return: the path
        """
        return self._path(root_dir, params)

    def build(self, root_dir: str, params: ArchetypeParameters) -> None:
        """
        Renders and saves the content generated by :py:meth:`render` to the location
        under `root_dir` determined by the _result of :py:meth:``subpath``.
        :param root_dir: the root directory under which the file should be saved
        :param params: the parameters used to determine the saved file content and
        possibly the subpath under the root directory
        :return: :py:const:`None`
        ..:note: this method has dependencies on the implementations of
        :py:meth:`render` and :py:meth:`subpath`
        """
        content = self.render(params)

        p = self._path(root_dir, params)
        try:
            os.makedirs(os.path.dirname(p))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

        with open(p, "w") as f:
            f.write(content)

    @abstractmethod
    def subpath(self, params: ArchetypeParameters) -> str:
        """
        Subclasses are required to implement this method.  Implementations of this
        method should return the subpath, under the root directory, of the file to be
        saved by :py:meth:`build`.
        :param params: the :py:class `ArchetypeParameters` to use as context when
        creating the subpath, e.g., when storing files whose sub-path might be
        determined by the package name.
        :return: the subpath
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR

    @abstractmethod
    def render(self, params: ArchetypeParameters) -> str:
        """
        Subclasses are required to implement this method.  Implementations of this
        method should return the content of the file to be saved by :py:meth:`build`.
        :param params: the :py:class:`ArchetypeParameters` to use as context when
        building the content
        :return: the content of the file to be saved by :py:meth:`build`
        """
        raise UNIMPLEMENTED_ABSTRACT_METHOD_ERROR



#!/usr/bin/env python
"""
    setup.py
    ~~~~~~~~

    Package distribution file for the inception_tools library.
"""

__author__ = 'Andrew van Herick'
__copyright__ = \
    'Unpublished Copyright (c) 2020 Andrew van Herick. All Rights Reserved.'
__license__ = 'Apache Software License 2.0'

import os

import setuptools

if __name__ == '__main__':
    # allow setup.py to run from another directory
    dir_ = os.path.abspath(os.path.join(__file__, os.pardir))
    os.chdir(dir_)
    setuptools.setup()
